// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnLinkedList.C   Version 1.22
//   LastSCCS    : 2/10/92  14:49:49
//   LastEdit    : "Mon Jan 20 13:53:00 1992"
//   Description : 
//   Author      : Andreas Baecker
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnDoubleLinkedList,GnObject,);

GnDoubleLinkedList::
GnDoubleLinkedList()
{
    InitEmpty();
}

GnDoubleLinkedList::
~GnDoubleLinkedList()
{
    Clear();
}

void GnDoubleLinkedList::
InitEmpty()
{
    head.next = head.previous = &head;
    length = 0;
    active_iterators = 0;
}

void GnDoubleLinkedList::
Clear()
{
    GnBiLinkable *link, *to_delete;

    for(link = head.next; link != &head; ) {
	to_delete = link;
	link = link->next;
	delete to_delete;
    }
    InitEmpty();
}

int GnDoubleLinkedList::
_Member(void *item) const 
{
    const GnBiLinkable *head_link = &head;
    
    for( GnBiLinkable *link = head.next; link != head_link; link = link->next)
	if( link->item == item )
	    return 1;
    return 0;
}

void *GnDoubleLinkedList::
_First() const
{    
    REQUIRE( ! Empty(), "List is not empty" );

    return head.next->item;
}

void *GnDoubleLinkedList::
_Last() const
{
    REQUIRE( ! Empty(), "List is not empty" );

    return head.previous->item;
}

void GnDoubleLinkedList::
_Append(void *item)
{
    register GnBiLinkable *new_link;

    new_link = new GnBiLinkable;
    new_link->item = item;
    new_link->next = &head;
    new_link->previous = head.previous;
    head.previous = new_link;
    new_link->previous->next = new_link;
    length++;
}

void GnDoubleLinkedList::
_Prepend(void *item)
{
    register GnBiLinkable *new_link;
    
    new_link = new GnBiLinkable;
    new_link->item = item;
    new_link->next = head.next;
    new_link->previous = &head;
    new_link->next->previous = new_link;
    head.next = new_link;
    length++;
}

void *GnDoubleLinkedList::
_RemoveFirst()
{
    REQUIRE( ! Empty(), "List is not empty" );

    register GnBiLinkable *to_delete;
    register void         *deleted_item;
    
    to_delete    = head.next;
    deleted_item = to_delete->item;
    head.next    = to_delete->next;
    head.next->previous = &head;
    length--;
    delete to_delete;
    return deleted_item;
}

void *GnDoubleLinkedList::
_RemoveLast()
{
    REQUIRE( ! Empty(), "List is not empty" );

    register GnBiLinkable *to_delete;
    register void         *deleted_item;
    
    to_delete     = head.previous;
    deleted_item  = to_delete->item;
    head.previous = to_delete->previous;
    head.previous->next = &head;
    length--;
    delete to_delete;
    return deleted_item;
}

void *GnDoubleLinkedList::
_Remove(void *deleted_item)
{
    register GnBiLinkable *to_delete;
    
    for(to_delete = head.next; to_delete != &head;
	to_delete = to_delete->next) {
	if( to_delete->item == deleted_item )  {
	    to_delete->previous->next = to_delete->next;
	    to_delete->next->previous = to_delete->previous;
	    length--;
	    delete to_delete;
	    return deleted_item;
	}
    }
}

void GnDoubleLinkedList::
_RemoveAllOccurences(void *deleted_item)
{
    register GnBiLinkable *ptr;
    register GnBiLinkable *to_delete;
    
    for(ptr = head.next; ptr != &head; ) {
	if( ptr->item == deleted_item )  {
	    to_delete = ptr;
	    ptr       = ptr->next;
	    to_delete->previous->next = to_delete->next;
	    to_delete->next->previous = to_delete->previous;
	    length--;
	    delete to_delete;
	} else {
	    ptr = ptr->next;
	}
    }
}
