/* @(#)GnPathname.C	1.20 7/19/91 */

META_IMPL_1(GnPathname,GnObject,);
ImplGenericList(GnPathnameList);

char *GnPathname::basename(char *path, char *suffix)
{
    // strips off the suffix (returns a copy if suffix == 0)

    Require(path != 0);
    
    if( ! suffix || ! strcmp(suffix, "") ) {
	return(strdup(path));
    } else {
	int path_len = strlen(path);
	int suffix_len = strlen(suffix);
	int base_len = path_len - suffix_len;
	if(base_len >= 0 && ! strcmp(suffix, path+base_len)) {
	    char *bname = new char[base_len+1];
	    strncpy(bname, path, base_len+1);
	    bname[base_len] = '\0';
	    return(bname);
	} else {
	    return(strdup(path));
	}
    }
}

char *GnPathname::dirname(char *)
{
    Require(0);
    return(0);
}

char *GnPathname::filename(char *)
{
    Require(0);
    return(0);
}

char *GnPathname::normalize_dir(char *)
{
    Require(0);
    return(0);
}


GnPathname ::
GnPathname()
{				/* Initialisiert mit der current working directory */
    dir = Gget_work_dir(&dirlen);
    nam = strdup("");
    typ = strdup("");
    namlen = typlen = 0;
    abs_name = rel_name = 0;
    make_abs_and_rel_names();

    Ensure( (dir != 0) && (dirlen > 0) && *dir == '/' );
}

GnPathname::GnPathname(char *name, char *type, char *directory, int do_stat)
{
    Require(((directory == 0) || (! strcmp(directory, "")) ||
	     (directory != 0 && strlen(directory) > 0 && *directory == '/')));

    if(! type) type = "";
    typlen = strlen(type);
    typ = new char[typlen+1];
    strcpy(typ, type);

    if( ! name ) name = "";
    nam = basename(name, typ);
    namlen = strlen(name);
    
    if(directory && strcmp(directory, "") ) {
	dirlen = strlen(directory);
	dir = new char[dirlen+1];
	strcpy(dir, directory);
    } else {
	dir = Gget_work_dir(&dirlen);
    }
    abs_name = rel_name = 0;
    make_abs_and_rel_names();
    
    Ensure( dir != 0 && strlen(dir) > 0 && *dir == '/' );
    Ensure( ( ! do_stat ) || ( exists() == Gstat_OK ) );
	
}

GnPathname::GnPathname(char *full_name, int do_stat)
{
    
    Require( (full_name != 0) );
    int fn_len = strlen(full_name);
    Require( fn_len > 0 );

    char *file_path;

    // Check for relative/absolute pathname
    if( *full_name == '/' ) {
	file_path = new char[fn_len+1];
	strcpy(file_path, full_name);
    } else {
	// relative path, prepend cwd();
	int wd_len;
	char *wd = Gget_work_dir(&wd_len);
	file_path = new char[wd_len+fn_len+2];
	strcpy(file_path, wd);
	if( strcmp(wd, "/") )
	    strcat(file_path, "/");
	strcat(file_path, full_name);
	delete wd;
    }
    
    // absolute pathname is now in 'file_path'

    char *fdir, *fname, *ftype;

    // Divide path into directory and filename
    fdir = strrchr(file_path, '/');
    if( fdir == file_path ) {
	dir = "/";
    } else {
	dir = file_path;
	*fdir = '\0';
    }
    fname = fdir+1;
 
    // Divide filename into name and extension
    if( *fname ) {
	ftype = strrchr(fname, '.');
	if( ftype == 0 ) {
	    ftype = "";
	} else {
	    char *rev;
	    // Test for a leading sequence of dots
	    for( rev = ftype; rev >= fname && *rev == '.'; rev-- );
	    if( rev == fname ) {
		ftype = "";
	    } else {
		char *t = ftype;
		ftype = new char[strlen(t)+1];
		strcpy(ftype, t);
		*t = '\0';
	    }
	}
    } else {
	fname = "";
	ftype = "";
    }
    
    dir = strdup(dir);
    dirlen = strlen(dir);
    nam = strdup(fname);
    namlen = strlen(nam);
    typ = strdup(ftype);
    typlen = strlen(typ);
    delete file_path;
    if( strcmp("", ftype) )
	delete ftype;

    abs_name = rel_name = 0;
    make_abs_and_rel_names();

    Ensure( dir != 0 && strlen(dir) > 0 && *dir == '/' );
    Ensure( ( ! do_stat ) || ( exists() == Gstat_OK ) );
}

GnPathname ::
~GnPathname()
{
    delete dir;
    delete nam;
    delete typ;
    delete abs_name;
    delete rel_name;
}

void GnPathname::make_abs_and_rel_names()
{
    if( abs_name ) delete abs_name;
    if( rel_name ) delete rel_name;
    abs_name = new char[dirlen+namlen+typlen+2];
    rel_name = new char[namlen+typlen+1];
    
    strcpy(rel_name, nam);
    strcat(rel_name, typ);
    
    strcpy(abs_name, dir);
    strcat(abs_name, "/");
    strcat(abs_name, rel_name);
}

Gstat GnPathname::
exists(char *p, struct stat* stat_buf_p)
{
    struct stat buf, *buf_ptr;

    buf_ptr = stat_buf_p ? stat_buf_p : &buf;
    
    if ( stat( p ? p : abs_name, buf_ptr) == 0 ) {
	Gina_Debug("buf.st_mtime = "); Gina_Debug_NL(buf_ptr->st_mtime);
	return Gstat_OK;
    } else {
	return(Gstat_NO);
    }
}

Gstat GnPathname::
do_stat(char *p, u_short mode_mask, u_short result_mask, u_short *prot)
{
    struct stat buf;

    if ( stat( p, &buf ) == 0 ) {
	if( prot != NULL )
	    *prot = buf.st_mode;
	if ( (buf.st_mode & mode_mask) == result_mask )
	    return Gstat_OK;
	else
	    return Gstat_NO;
    }
    return(Gstat_FAILED);
}

Gstat GnPathname::is_regular_file(u_short *prot)
{
    return(do_stat(abs_name, S_IFREG, S_IFREG, prot));
}

Gstat GnPathname::is_directory(u_short *prot)
{
    return(do_stat(abs_name, S_IFDIR, S_IFDIR, prot));
}

Gstat GnPathname::directory_valid(u_short *prot)
{
    return(do_stat(dir, S_IFDIR, S_IFDIR, prot));
}

Gstat GnPathname::writeable()
{
    int fd = open(abs_name, O_WRONLY) ;

    if( fd == -1 )
	return Gstat_NO;
    close(fd);
    unlink(abs_name);
    return Gstat_OK;
}

Gstat GnPathname::writeable_or_createable()
{
    int fd = open(abs_name, O_WRONLY | O_CREAT, 0777) ;
    
    if( fd == -1 )
	return Gstat_NO;
    close(fd);
    unlink(abs_name);
    return Gstat_OK;
}

Gstat GnPathname::readable()
{
    int fd = open(abs_name, O_RDONLY) ;

    if( fd == -1 )
	return Gstat_NO;
    close(fd);
    return Gstat_OK;
}

int GnPathname::open_fd(int flags, int mode)
{
    return(open(abs_name, flags, mode));
}

GnPathname ::
operator char*()
{				/* konvertiert pathname nach String */
    return abs_name;
}

char *GnPathname::pattern()
{
    char *pat = new char[typlen+3];
    pat[0] = '?';
    pat[1] = '*';
    strcpy(pat+2, typ);
    return pat;
}

void GnPathname ::
set_directory( char *directory )
{
    Require( (directory != 0) && (strlen( directory ) > 0) );

    delete dir;
    dirlen = strlen(directory);
    dir = new char[dirlen+1];
    strcpy(dir, directory);
    if(dirlen > 1 && dir[dirlen-1] == '/') {
	--dirlen;
	dir[dirlen] = '\0';
    }
    make_abs_and_rel_names();
}

void GnPathname ::
set_name( char *name )
{
    Require( (name != 0) && (strlen( name ) > 0) );
    delete nam;
    namlen = strlen(name);
    nam = new char[namlen+1];
    strcpy(nam, name);
    make_abs_and_rel_names();
}

void GnPathname ::
set_type( char *type )
{
    Require( (type != 0) && (strlen( type ) > 0) );
    delete typ;
    typlen = strlen(type);
    typ = new char[typlen+1];
    strcpy(typ, type);
    make_abs_and_rel_names();
}

void GnPathname ::
merge_dir( char *p )
{
    Require( p != 0 && strlen( p ) > 0 );
    Require( (dir != 0 && *dir == '/') ? ( *p != '/') : 1 );
    Require(0);
}

int GnPathname ::
select( u_short ent_typ )
{
    struct stat buf;

    if ( stat( abs_name, &buf ) == 0 )
	if ( (buf.st_mode & S_IFMT) == ent_typ )
	    return 1;
    return 0;
}

int GnPathname ::
select_dirs()
{
    return select( S_IFDIR );
}

int GnPathname ::
select_files()
{
    return select( S_IFREG );
}

int GnPathname ::
comp( GnPathname *pat )
{
    int erg = strcmp( abs_name, pat->abs_namestring() );
    return erg;
}

//GnPathnameList * GnPathname ::
//scan_dir( int (GnPathname::*select_fct)(), int (GnPathname::*comp_fct)(GnPathname *) )
//{
    //DIR *dirp = opendir( dir );
    //if ( dirp == 0 ) return 0;
    //GnPathnameList *res = new GnPathnameList;
    //GnPathname *p;
    //for ( struct dirent *dp = readdir( dirp ); dp != NULL; dp = readdir( dirp ) ) {
	//p = new GnPathname( dp->d_name, NULL, dir );
	//if ( (! select_fct) ) // || (p->select_fct()) )
	    //res->append( p );
    //else
	//delete p;
    //}
    //closedir( dirp );
    //if ( comp_fct );
    // Klasse List braucht noch ein Sort, oder NIHCL !!!
    // res->sort( comp_fct );
    //return res;
//}

//GnPathnameList * GnPathname ::
//list_all_names_unsort()
//{
    //return scan_dir();
//}

//GnPathnameList * GnPathname ::
//list_all_names()
//{
    //return scan_dir( NULL, &GnPathname::comp );
//}

//GnPathnameList * GnPathname ::
//list_sub_directories()
//{
    //return scan_dir( &GnPathname::select_dirs, &GnPathname::comp );
//}

//GnPathnameList * GnPathname ::
//list_files()
//{
    //return scan_dir( &GnPathname::select_files, &GnPathname::comp );
//}

char *
Gget_work_dir(int *len)
{				/* liefert die current working directory */
    char path[ MAXPATHLEN ];
    char *p = getwd( path );
    int dirlength = strlen(p);
    char *result = new char [ dirlength + 1 ];

    strcpy( result, p );
    if(len)
	*len = dirlength;
    return result;
}
