// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnCallback.h   Version 1.23
//   LastSCCS    : 1/23/92  12:32:13
//   LastEdit    : "Mon Jan 20 13:09:49 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

#ifndef Gina_Callback_H
#define Gina_Callback_H

class GnTimer;

typedef void (*GinaSimpleCallbackProc) (caddr_t);

class GnCallback : public GnObject {
    META_DEF_1(GnCallback,GnObject);
    friend class GnTimer;
  public:
    enum GinaCallbackType { CB_NOT_ATTACHED, CB_WIDGET, CB_EVENT,
			    CB_PROTOCOL, CB_TIMER};
  public:
    GnCallback(GinaSimpleCallbackProc);
    ~GnCallback();
    
    virtual void evaluate(caddr_t call_data);
    void attach_to_widget(Widget, char *);
    void attach_to_widget(Widget, EventMask mask, Boolean non_maskable);
    void attach_to_widget(Widget, Atom, Atom);
    void attach_to_timer(GnTimer *);
    void remove();
    
  protected:

    // Callback handlers (called from "C" only)
    static void widget_handler(Widget w, GnCallback* client_data,
			       caddr_t call_data);
    static void timer_handler(GnTimer *, XtIntervalId *);
    
    GinaSimpleCallbackProc function_name;
    GinaCallbackType cb_type;
    union {
	struct {Widget widget; char *name; } callback;
	struct {Widget widget; EventMask mask; Boolean non_maskable;} event;
	struct {Widget widget; Atom property, protocol; } atom;
	struct {GnTimer *timer; } timer;
    } cb_info;
};

DefGenericListWithIterator(GnCallbackList,GnCallback,GnCallbackListIterator);

typedef void (*GinaClassCallbackProc) (void *, caddr_t);

class GnClassCallback : public GnCallback {
    META_DEF_1(GnClassCallback,GnCallback);
  public:
    GnClassCallback(GinaClassCallbackProc proc, void* object);
    virtual void evaluate(caddr_t call_data);
  protected:
    void *object;
};

typedef void (*GinaClassCallbackProcOne) (void *, void *, caddr_t);

class GnCallbackOne : public GnClassCallback {
     META_DEF_1(GnCallbackOne,GnClassCallback);
  public:
    GnCallbackOne(GinaClassCallbackProcOne proc, void* object, void *);
    virtual void evaluate(caddr_t call_data);
  protected:
    void *client_data_one;
};

typedef void (*GinaClassCallbackProcTwo) (void *, void *, void *, caddr_t);

class GnCallbackTwo : public GnCallbackOne {
     META_DEF_1(GnCallbackTwo,GnCallbackOne);
  public:
    GnCallbackTwo(GinaClassCallbackProcTwo proc, void*, void *, void *);
    virtual void evaluate(caddr_t call_data);
  protected:
    void *client_data_two;
};

#define SIMPLE_CALLBACK(function)  (new GnCallback((function)))

#define SIMPLE_CALLBACK1(function, client_data_class, client_data) \
           (new GnClassCallback((GinaClassCallbackProc)(function), \
		               (void *)(client_data_class)(client_data)))

#define SIMPLE_CALLBACK2(function,cld_class_1,cld_1,cld_class_two,cld_two) \
          (new GnCallbackOne((GinaClassCallbackProcOne)(function), \
                            (void *)(cld_class_1)(cld_1),\
			    (void *)(cld_class_two)(cld_two)))

#define CALLBACK(class, function, object) \
         (new GnClassCallback((GinaClassCallbackProc)(class :: function), \
			     (void *)(class *)(object)))

#define CALLBACK1(class,function,object,client_data_class_1,client_data_1) \
         (new GnCallbackOne((GinaClassCallbackProcOne)(class :: function), \
			   (void *)(class *)(object), \
			   (void *)(client_data_class_1)(client_data_1)))

#define CALLBACK2(class,function,object,cld_class_1,cld_1,cld_class_two,cld_2) \
         (new GnCallbackTwo((GinaClassCallbackProcTwo)(class :: function), \
                           (void *)(class *)(object), \
			   (void *)(cld_class_1)(cld_1), \
			   (void *)(cld_class_two)(cld_2)))

#endif Gina_Callback_H
