// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnContracts.h   Version 1.3
//   LastSCCS    : 1/23/92  12:32:37
//   LastEdit    : "Mon Jan 20 14:20:25 1992"
//   Description : Programming by contract in C++
//   Author      : Andreas Baecker
//   Copyright   : GMD Schloss Birlinghoven

#ifndef _Gncontracts_h_
#define _Gncontracts_h_

class GnContracts {
  public:
    friend void RequireCondition(int p_condition, char *msg,
				 char *p_file, int p_line);
    friend void EnsureCondition(int p_condition, char *msg,
				char *p_file, int p_line);
    friend void CheckCondition(int p_condition, char *msg,
			       char *p_file, int p_line);
    friend void CheckInvariant(GnContracts *p_object,
			       char *p_file, int p_line);
    friend void CheckInvariant(GnContracts &p_object,
			       char *p_file, int p_line);
  public:
    virtual int ClassInvariant();
};

#if defined(CONTRACTS) || defined(GINA_DEBUG)

#  define REQUIRE(_EXP_,_COMMENT_) \
      RequireCondition(_EXP_, _COMMENT_, __FILE__, __LINE__)

#  define ENSURE(_EXP_, _COMMENT_) \
      EnsureCondition(_EXP_, _COMMENT_, __FILE__, __LINE__)

#  define REQUIRE_IF(_LEFT_, _RIGHT_, _COMMENT_) \
      if(_LEFT_) REQUIRE(_RIGHT_,_COMMENT_)

#  define ENSURE_IF(_LEFT_, _RIGHT_, _COMMENT_) \
      if(_LEFT_) ENSURE(_RIGHT_,_COMMENT_)

#  define REMEMBER(_TYPE_,_VAR_) \
      const _TYPE_ _NAME2_(ContractOld,_VAR_) = _VAR_

#  define REMEMBER_VALUE(_TYPE_,_VAR_,_EXPRESSION_) \
      const _TYPE_ _NAME2_(ContractOldValue,_VAR_) = (_EXPRESSION_)

#  define OLD(_VAR_) \
      _NAME2_(ContractOld,_VAR_)

#  define OLD_VALUE(_VAR_) \
      _NAME2_(ContractOldValue,_VAR_)

#  define CHECK(_EXP_,_COMMENT_) \
      CheckCondition(_EXP_, _COMMENT_, __FILE__, __LINE__)

#  define CHECK_INVARIANT() \
      CheckInvariant(this)

#  define CHECK_OBJECT(_OBJECT_) \
      CheckInvariant((_OBJECT_))
#else
#  define REQUIRE(_EXP_,_COMMENT_)
#  define ENSURE(_EXP_, _COMMENT_)
#  define REQUIRE_IF(_LEFT_, _RIGHT_, _COMMENT_)
#  define ENSURE_IF(_LEFT_, _RIGHT_, _COMMENT_)
#  define REMEMBER(_TYPE_,_VAR_)
#  define REMEMBER_VALUE(_TYPE_,_VAR_,_EXPRESSION_)
#  define OLD(_VAR_)
#  define OLD_VALUE(_VAR_)
#  define CHECK(_EXP_,_COMMENT_)
#  define CHECK_INVARIANT()
#  define CHECK_OBJECT(_OBJECT_)
#endif

#if defined(CONTRACTS) || defined(GINA_DEBUG)
    /* EXIT may be defined as abort() or exit(<state>) */
#   define EXIT abort()

#   define Require(ex) \
      {if(!(ex)){cerr<<"Precondition violated: file "<<__FILE__<< \
      ", line "<<__LINE__<<endl;EXIT;}}

#   define Ensure(ex) \
      {if(!(ex)){cerr<<"Postcondition violated: file "<<__FILE__<< \
      ", line "<<__LINE__<<endl;EXIT;}}

#   define Invariant(ex) \
      {if(!(ex)){cerr<<"Invariant violated: file "<<__FILE__<< \
       ", line "<<__LINE__<<", Method: "<<method<<endl;EXIT;}}

#   define Check_Invariant(me)	{invariant(me);}

#else
#   define Require(ex)
#   define Ensure(ex)
#   define Invariant(ex)
#   define Check_Invariant(me)
#endif

#endif
