// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnMacros.h   Version 1.13
//   LastSCCS    : 1/14/92  15:15:56
//   LastEdit    : "Fri Nov 29 13:37:27 1991"
//   Description : 
//   Author      : Andreas Baecker
//   Copyright   : GMD Schloss Birlinghoven

#ifndef GnMacros.h
#define GnMacros.h

#define GnXMSTRING_LTOR(s) \
    XmStringCreateLtoR(s, GnWidget::TheToolkit()->AppCharSet())

// Define a resource with accessor and modifier functions

#define DEF_SET_GET(r_name, r_type) \
    void _NAME2_(setR_,r_name)(r_type value); \
    r_type _NAME2_(getR_,r_name)()

// Define a resource with accessor and modifier functions as inline functions

#define DEF_SET_GET_I(r_name, r_type) \
    void _NAME2_(setR_,r_name)(const r_type value) \
      { set_arg(_NAME2_(XmN,r_name), (XtArgVal)value); }; \
    r_type _NAME2_(getR_,r_name)() \
      { r_type v; get_arg(_QUOTE_(r_name), (XtArgVal)&v); return v; }

// Define a resource with a modifier function

#define DEF_SET(r_name, r_type) \
    void _NAME2_(setR_,r_name)(r_type value)

// Define a resource with a modifier function as an inline function

#define DEF_SET_I(r_name, r_type) \
    void _NAME2_(setR_,r_name)(const r_type value) \
      { set_arg(_NAME2_(XmN,r_name), (XtArgVal)value); };

// Define a resource with an accessor function

#define DEF_GET(r_name, r_type) r_type _NAME2_(getR_,r_name)()

// Define a resource with an accessor function as an inline function

#define DEF_GET_I(r_name, r_type) \
    r_type _NAME2_(getR_,r_name)() \
      { r_type v; get_arg(_QUOTE_(r_name), (XtArgVal)&v); return v; }

// Define a XmString resource with accessor and modifier and an automtic
// conversion from char * to XmString (relies on XmString being defined
// as unsigned char *)

#define DEF_XMSTRING(r_name)            \
    DEF_SET_GET_I(r_name, XmString);    \
    DEF_SET(r_name, char *)             \
      { set_xmstring_arg(_NAME2_(XmN,r_name), value); }

// Define a pixmap resource with accessor and modifier

#define DEF_PIXMAP(r_name) \
    void _NAME2_(setR_,r_name)(const Pixmap value) \
      { set_arg(_NAME2_(XmN,r_name), (XtArgVal)value); }; \
    void _NAME2_(setR_,r_name)(char *const value) \
      { set_pixmap_arg(_NAME2_(XmN,r_name), value); }; \
    void _NAME2_(setR_,r_name)(char *value, unsigned int width, \
                                    unsigned int height) \
      { set_pixmap_arg(_NAME2_(XmN,r_name), value, width, height); } \
    Pixmap _NAME2_(getR_,r_name)() \
      { Pixmap v; get_arg(_QUOTE_(r_name), (XtArgVal)&v); return v; }

// Define a fontList resource with accessor and modifier

#define DEF_FONTLIST(r_name) \
    void _NAME2_(setR_,r_name)(const XmFontList value) \
      { set_arg(_NAME2_(XmN,r_name), (XtArgVal)value); }; \
    void _NAME2_(setR_,r_name)(char *const value) \
      { set_font_arg(_NAME2_(XmN,r_name), value); }; \
    XmFontList _NAME2_(getR_,r_name)() \
      { XmFontList v; get_arg(_QUOTE_(r_name), (XtArgVal)&v); return v; }

// Implementation for a GnWidget* resource

#define WIDGET_RESOURCE_DEF(r_name,r_type) \
    DEF_SET_GET(r_name,r_type *); \
    DEF_SET(r_name, r_type &)

#define WIDGET_RESOURCE_IMPL(w_class, r_class, r_name) \
    void w_class:: \
    _NAME2_(setR_,r_name)(r_class *tmp) \
    { \
        set_arg(_NAME2_(XmN,r_name), (XtArgVal)(tmp->get_widget_id()));  \
    } \
 \
    void w_class:: \
    _NAME2_(setR_,r_name)(r_class &tmp) \
    { \
        set_arg(_NAME2_(XmN,r_name), (XtArgVal)(tmp.get_widget_id()));  \
    } \
 \
    r_class *w_class::_NAME2_(getR_,r_name)() \
    { \
        register Widget   tmp_widget; \
        register GnWidget *tmp; \
 \
        get_arg(_NAME2_(XmN,r_name), (XtArgVal)&tmp_widget); \
        if( tmp_widget ) { \
	    tmp = WidgetTable()->lookup(tmp_widget); \
	    if( ! tmp ) \
                tmp = WidgetTable()->lookup(tmp_widget,\
                                                   new r_class, True); \
	    return tmp ? r_class::TypeSafeDownCast((GnObject *)tmp) : 0; \
        } else { \
	    return 0; \
        } \
    }

// Define a constraint with accessor and modifier

#define C_DEF_SET_GET(c_name, c_type) \
    void _NAME2_(setC_,c_name)(GnWidget &w, c_type a); \
    c_type _NAME2_(getC_,c_name)(GnWidget &w)

// Define a constraint with accessor and modifier as inline functions

#define C_DEF_SET_GET_I(c_name, c_type) \
    void _NAME2_(setC_,c_name)(GnWidget &w, c_type a) \
	{ w.get_group_widget()->set_arg(_NAME2_(XmN,c_name), \
					(XtArgVal)a); }; \
    c_type _NAME2_(getC_,c_name)(GnWidget &w) \
	{ c_type v; \
          w.get_group_widget()->get_arg(_NAME2_(XmN,c_name), \
					(XtArgVal)&v); \
          return(v);}

// Define a constraint with modifier

#define C_DEF_SET(c_name, c_type) \
    void _NAME2_(setC_,c_name)(GnWidget &w, c_type a)

#define CALLBACK_DEF(c_name) \
  void _NAME2_(add_,c_name)(GnCallback *cb) \
    { add_callback(cb, _NAME2_(XmN,c_name)); }; \
  void _NAME2_(remove_,c_name)(GnCallback *cb, int free_it = 1 ) \
      { remove_callback(cb, free_it); }

#ifdef USE_NIHCL
#else

# ifdef ABS
# undef ABS
# endif

inline char	ABS(char x)	{ return x >= 0 ? x : -x; }
inline short	ABS(short x)	{ return x >= 0 ? x : -x; }
inline int	ABS(int x)	{ return x >= 0 ? x : -x; }
inline long	ABS(long x)	{ return x >= 0 ? x : -x; }
inline float	ABS(float x)	{ return x >= 0 ? x : -x; }
inline double	ABS(double x)	{ return x >= 0 ? x : -x; }

#ifdef MAX
#undef MAX
#endif

inline char	MAX(char a,char b)	{ return a >= b ? a : b; }
inline short	MAX(short a,short b)	{ return a >= b ? a : b; }
inline int	MAX(int a,int b)	{ return a >= b ? a : b; }
inline long	MAX(long a,long b)	{ return a >= b ? a : b; }
inline float	MAX(float a,float b)	{ return a >= b ? a : b; }
inline double	MAX(double a,double b)	{ return a >= b ? a : b; }
inline void*	MAX(void* a,void* b)	{ return a >= b ? a : b; }
inline unsigned char	MAX(unsigned char a, unsigned char b)
       { return a >= b ? a : b; }
inline unsigned short	MAX(unsigned short a, unsigned short b)
       { return a >= b ? a : b; }
inline unsigned int	MAX(unsigned int a, unsigned int b)
       { return a >= b ? a : b; }
inline unsigned long	MAX(unsigned long a, unsigned long b)
       { return a >= b ? a : b; }

#ifdef MIN
#undef MIN
#endif

inline char	MIN(char a,char b)	{ return a <= b ? a : b; }
inline short	MIN(short a,short b)	{ return a <= b ? a : b; }
inline int	MIN(int a,int b)	{ return a <= b ? a : b; }
inline long	MIN(long a,long b)	{ return a <= b ? a : b; }
inline float	MIN(float a,float b)	{ return a <= b ? a : b; }
inline double	MIN(double a,double b)	{ return a <= b ? a : b; }
inline void*	MIN(void* a,void* b)	{ return a <= b ? a : b; }
inline unsigned char	MIN(unsigned char a, unsigned char b)
       { return a <= b ? a : b; }
inline unsigned short	MIN(unsigned short a, unsigned short b)
       { return a <= b ? a : b; }
inline unsigned int	MIN(unsigned int a, unsigned int b)
       { return a <= b ? a : b; }
inline unsigned long	MIN(unsigned long a, unsigned long b)
       { return a <= b ? a : b; }
#endif

#endif
