// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnMetaClass.h   Version 1.8
//   LastSCCS    : 1/23/92  12:32:33
//   LastEdit    : "Mon Jan 20 10:36:56 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

#ifndef Gnmetaclass_h
#define Gnmetaclass_h

// Many of the ideas here are taken from ET++ and NIHCL
// New features are:
//  - Support for multiple inheritance
//  - Support virtual base classes and abstract virtual base classes
//

/*
   Problems:
   - Prototype objects with virtual base classes could not be created as
     arguments for the static class objects. The compiler complains about:
     "sorry, not implemented: default argument too complicated"
     Solution: Use the function CreatePrototype() to create the object.
   - Prototype objects of abstract classes could not be created.
     So we threw out CreatePrototype() and pass the address of
     THIS_CLASS::GetBaseClasses to the meta class object's constructor.
   - We first made GetBaseClasses() a static function. This caused 
     cfront to die with an "internal error". So we made it a true member
     function, added some type casts and call the function with a NULL
     this-pointer in GnClass::Initialize()
*/

class GnClassInfoList;
class GnClassMemberList;

typedef void (*GnBaseClassCollector)(class GnObject *,
				     class GnClassInfoList *);
				     
typedef void (*GnClassMemberCollector)(class GnObject *,
				       class GnClassMemberList *);

// Define metaclass information for simple (non-class) types

#define _SimpleMetaDef(FULL_TYPE,THIS_TYPE,VAR_PREFIX) \
  extern class GnClass *_NAME2_(VAR_PREFIX,THIS_TYPE)ImplPtr; \
  inline class GnClass *TypeOf(const FULL_TYPE *) \
      { return _NAME2_(VAR_PREFIX,THIS_TYPE)ImplPtr; }

#define SimpleMetaDef(THIS_TYPE) \
    _SimpleMetaDef(THIS_TYPE,THIS_TYPE,class_of_)
#define SimpleMetaDefUnsigned(THIS_TYPE) \
    _SimpleMetaDef(unsigned THIS_TYPE,THIS_TYPE,class_of_u_)

SimpleMetaDef(char)
SimpleMetaDef(int)
SimpleMetaDef(short)
SimpleMetaDef(long)
SimpleMetaDef(float)
SimpleMetaDef(double)
SimpleMetaDefUnsigned(char)
SimpleMetaDefUnsigned(short)
SimpleMetaDefUnsigned(int)
SimpleMetaDefUnsigned(long)

#define _SimpleMetaImpl(FULL_TYPE,THIS_TYPE,VAR_PREFIX,REP_TYPE) \
    class _NAME2_(VAR_PREFIX,THIS_TYPE) : public GnClass { \
      public: \
        virtual int IsSimpleType() \
	    { return 1; } \
        virtual int IsClassType() \
	    { return 0; } \
        _NAME2_(VAR_PREFIX,THIS_TYPE)() \
            : GnClass( _QUOTE_(FULL_TYPE), sizeof(FULL_TYPE), \
                __FILE__, __LINE__, __FILE__, __LINE__, 0, 0) {} \
        virtual int StoreOn(ostream &ostr, char *addr); \
        virtual int ReadFrom(istream &istr, char *addr); \
    }; \
    _NAME2_(VAR_PREFIX,THIS_TYPE) _NAME2_(VAR_PREFIX,THIS_TYPE)Impl;\
    GnClass *_NAME2_(VAR_PREFIX,THIS_TYPE)ImplPtr = \
        &_NAME2_(VAR_PREFIX,THIS_TYPE)Impl;\
    int _NAME2_(VAR_PREFIX,THIS_TYPE):: \
    StoreOn(ostream &ostr, char *addr) \
    { \
        ostr << (REP_TYPE)(*(FULL_TYPE *)addr) << endl; \
        return ostr.fail() ? 0 : 1; \
    } \
    int _NAME2_(VAR_PREFIX,THIS_TYPE):: \
    ReadFrom(istream &istr, char *addr) \
    { \
      REP_TYPE buf; \
      istr >> buf; \
      *(FULL_TYPE *)addr = (FULL_TYPE)buf; \
      return istr.fail() ? 0 : 1; \
    }
    
#define SimpleMetaImpl(THIS_TYPE) \
    _SimpleMetaImpl(THIS_TYPE,THIS_TYPE,ClassOf,THIS_TYPE)
#define SimpleMetaImplUnsigned(THIS_TYPE) \
    _SimpleMetaImpl(unsigned THIS_TYPE,THIS_TYPE,ClassOfUnsigned,unsigned THIS_TYPE)

// Define metaclass information and downcast operators

#define _META_DEF_0(THIS_CLASS)                                             \
  public:                                                                   \
    THIS_CLASS(class GnDummy &d );                                          \
    static GnClass *ClassOf()                                               \
	{ return(class_of); }                                               \
    virtual GnClass *Class();                                               \
    static int KindOf(GnObject *obj)					    \
        { return obj->Class() == class_of; }				    \
    static int KindOf(GnObject &obj)					    \
        { return obj.Class() == class_of; }				    \
  public:                                                                   \
    void GetBaseClasses(GnClassInfoList *);                                 \
    void GetClassMembers(GnClassMemberList *);                              \
    static char *GetDeclFileName()                                          \
        { return __FILE__; }                                                \
    static int GetDeclFileLine()                                            \
        { return __LINE__; }                                                \
    static char *GetImplFileName();                                         \
    static int GetImplFileLine();                                           \
    friend class GnClass *TypeOf(THIS_CLASS *)                              \
        { return THIS_CLASS::class_of; }                                    \
  public:								    \
    virtual void *_TypeSafeDownCast(GnClass *);      	  		    \
    virtual void *_GetThisPointer();                                        \
    static THIS_CLASS *TypeSafeDownCast(GnObject *o)                        \
	{ return (THIS_CLASS *)o->_TypeSafeDownCast(class_of); }            \
    static THIS_CLASS &TypeSafeDownCast(GnObject &o)                        \
	{ return *(THIS_CLASS *)o._TypeSafeDownCast(class_of); }            \
  protected:                                                                \
    static GnClass *class_of

#define META_DEF_0(THIS_CLASS) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_1(THIS_CLASS,BASE_CLASS) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_1_V(THIS_CLASS,BASE_CLASS) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_2(THIS_CLASS,BASE_CLASS_1,BASE_CLASS_2) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_2_VN(THIS_CLASS,BASE_CLASS_1,BASE_CLASS_2) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_2_NV(THIS_CLASS,BASE_CLASS_1,BASE_CLASS_2) \
    _META_DEF_0(THIS_CLASS)

#define META_DEF_2_VV(THIS_CLASS,BASE_CLASS_1,BASE_CLASS_2) \
    _META_DEF_0(THIS_CLASS)

// ============================================================================

// Implementation of features which only depend on the class itself

#define _META_IMPL_0(THIS_CLASS) \
  void *THIS_CLASS::_GetThisPointer() \
      { return (void *)(THIS_CLASS *)this; } \
  GnClass *THIS_CLASS::Class()                                              \
	{ return(class_of); }                                               \
  char *THIS_CLASS::GetImplFileName() \
      { return __FILE__; } \
  int THIS_CLASS::GetImplFileLine() \
      { return __LINE__; } \
  GnClass *THIS_CLASS::class_of = \
      new GnClass(_QUOTE_(THIS_CLASS), \
                  sizeof(THIS_CLASS), \
		  THIS_CLASS::GetDeclFileName(), \
                  THIS_CLASS::GetDeclFileLine(), \
                  THIS_CLASS::GetImplFileName(), \
                  THIS_CLASS::GetImplFileLine(), \
                  (GnBaseClassCollector)(THIS_CLASS::GetBaseClasses), \
                  (GnClassMemberCollector)(THIS_CLASS::GetClassMembers));

// Implementation of features for a class without base classes

#define META_IMPL_0(THIS_CLASS) \
  _META_IMPL_0(THIS_CLASS) \
  THIS_CLASS::THIS_CLASS(class GnDummy &) {} \
  void THIS_CLASS::GetBaseClasses(GnClassInfoList *) {} \
  void THIS_CLASS::GetClassMembers(GnClassMemberList *) {} \
  void *THIS_CLASS::_TypeSafeDownCast(GnClass *c) \
  { \
      if( c == class_of ) return (void *)this; \
      return 0; \
  }

//============================================================================

// Implementation of features for a class with 1 base class

#define _META_IMPL_1(THIS_CLASS,BASE_CLASS,V_PREFIX,VIRTUAL,MEMBERS) \
  THIS_CLASS::THIS_CLASS(class GnDummy &d) : BASE_CLASS(d) {} \
  void THIS_CLASS::GetBaseClasses(GnClassInfoList *cl) \
  { \
      cl->Append(new GnClassInfo(BASE_CLASS::ClassOf(),VIRTUAL)); \
  } \
  void THIS_CLASS::GetClassMembers(GnClassMemberList *) \
  { \
      MEMBERS \
  } \
  void *THIS_CLASS::_TypeSafeDownCast(GnClass *c) \
  { \
      if( c == class_of ) return (void *)this; \
      return BASE_CLASS::_TypeSafeDownCast(c); \
    }

#define META_IMPL_1(THIS_CLASS,BASE_CLASS,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_1(THIS_CLASS,BASE_CLASS,,0,MEMBERS)

#define META_IMPL_1_V(THIS_CLASS,BASE_CLASS,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_1(THIS_CLASS,BASE_CLASS,Virtual,1,MEMBERS)

//============================================================================

// Implementation of features for a class with 2 base classes

#define _META_IMPL_2(THIS_CLASS,B_C_1,VP_1,V_1,B_C_2,VP_2,V_2,MEMBERS) \
  THIS_CLASS::THIS_CLASS(class GnDummy &d) \
      : B_C_1(d),B_C_2(d) {} \
  void THIS_CLASS::GetBaseClasses(GnClassInfoList *cl) \
  { \
      cl->Append(new GnClassInfo(B_C_1::ClassOf(),V_1)); \
      cl->Append(new GnClassInfo(B_C_2::ClassOf(),V_2)); \
  } \
  void THIS_CLASS::GetClassMembers(GnClassMemberList *) \
  { \
      MEMBERS \
  } \
  void *THIS_CLASS::_TypeSafeDownCast(GnClass *c) \
  { \
      if( c == class_of ) return (void *)this; \
      return c->AmbiguousCast(B_C_1::_TypeSafeDownCast(c), \
                              B_C_2::_TypeSafeDownCast(c), \
                              class_of); \
  } 

#define META_IMPL_2(THIS_CLASS,B_C_1,B_C_2,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_2(THIS_CLASS,B_C_1,,0,B_C_2,,0,MEMBERS)

#define META_IMPL_2_VN(THIS_CLASS,B_C_1,B_C_2,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_2(THIS_CLASS,B_C_1,Virtual,1,B_C_2,,0,MEMBERS)

#define META_IMPL_2_NV(THIS_CLASS,B_C_1,B_C_2,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_2(THIS_CLASS,B_C_1,,0,B_C_2,Virtual,1,MEMBERS)

#define META_IMPL_2_VV(THIS_CLASS,B_C_1,B_C_2,MEMBERS) \
  _META_IMPL_0(THIS_CLASS) \
  _META_IMPL_2(THIS_CLASS,B_C_1,Virtual,1,B_C_2,Virtual,1,MEMBERS)

#endif








