// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnSelection.h   Version 1.9
//   LastSCCS    : 1/23/92  12:32:26
//   LastEdit    : "Wed Jan 22 14:35:03 1992"
//   Description : 
//   Author      : Andreas Baecker
//   Copyright   : GMD Schloss Birlinghoven

#ifndef GnSelection_h
#define GnSelection_h

class GnSelection : public GnObject {
    META_DEF_1(GnSelection,GnObject);
  public:
    GnSelection();
    
    // Add an item to the item list
    GnPresentable *add(GnPresentable *, Boolean = False, Boolean = True);
    GnPresentable *add(GnPresentable &, Boolean = False, Boolean = True);
    GnPresentable *add(char *, int, Boolean = False, Boolean = True);
    GnPresentable *add(char *, float, Boolean = False, Boolean = True);
    GnPresentable *add(char *, Boolean = False, Boolean = True);
    GnPresentable *add(int, Boolean = False, Boolean = True);
    GnPresentable *add(float, Boolean = False, Boolean = True);
    void add(GnPresentableList *, GnCallback *callback = 0);
    void add(GnPresentable *item_list[], int item_count,
	     GnCallback *callback = 0);

    // Remove an Item from the list
    GnPresentable *Remove(GnPresentable *p_item);

    // Insert an item into the list
    GnPresentable *Insert(GnPresentable *p_item, GnPresentable *p_before,
			  Boolean p_selected = False,
			  Boolean p_sensitive = True);
    
    // Manipulate the item list
    void SetItemList(GnPresentableList *new_item_list);
    void SetItemList(GnPresentable *new_item_list[], int new_item_count);

    // New callbacks
    void add_selectionCallback(GnCallback *);
    void remove_selectionCallback(GnCallback *, Boolean free_it = True);

    // Access to items. Clients must free the temporary lists !!
    GnPresentableList *AllItems();
    GnPresentableList *SelectedItems();
    GnPresentableList *UnselectedItems();
    GnPresentableList *SensitiveItems();
    GnPresentableList *InsensitiveItems();

    // return the Object's position in the list (first = 0)
    int Position(GnPresentable *);

    // Test whether an Object is currently selected
    Boolean IsSelected(GnPresentable *);
    Boolean IsSelected(GnPresentable &);
    // Test whether an Object is currently sensitive
    Boolean IsSensitive(GnPresentable *);
    Boolean IsSensitive(GnPresentable &);

    // Change selection
    void Select(GnPresentable *, Boolean notify = False);
    void Unselect(GnPresentable *, Boolean notify = False);

    // Change sensitivity
    void Sensitive(GnPresentable *);
    void Insensitive(GnPresentable *);
    
    // Count the number of selected items
    int NumberOfSelectedItems(GnSelectionItem **p_first_selected_item = 0);

    // Count the number of sensitive items
    int NumberOfSensitiveItems(GnSelectionItem **p_first_sensitive_item = 0);

  protected:
    // Find the selection item corresponding to a presentable object
    GnSelectionItem *FindSelectionItem(GnPresentable *p_item);

    // True if each item has it's own widget 
    virtual Boolean HasWidgets()
	{ return False; }
    virtual Boolean AllowChangeSensitivity()
	{ return True; }

    // Create the item for an object in the selection
    virtual GnSelectionItem *create_item(GnPresentable *, Boolean, Boolean);
    virtual void Removed(GnSelectionItem *p_item);

    virtual void VisualizeItem(GnSelectionItem *);
    virtual void UnvisualizeItem(GnSelectionItem *);
    virtual void VisualizeAllItems();
    virtual void UnvisualizeAllItems();
    
    virtual void SelectItem(GnSelectionItem *, Boolean, Boolean);
    virtual void SelectItemWidget(GnSelectionItem *, Boolean, Boolean);
    virtual void SensitiveItemWidget(GnSelectionItem *, Boolean);
    virtual void SensitiveItem(GnSelectionItem *, Boolean);
    
    virtual void CheckSelectedItems();

    // Used for mapping callbacks
    void change_value(void *client_data, caddr_t call_data);
    virtual void call_selection_callback();
    virtual void selection_changed();
  protected:
    GnSelectionItemList entries;
    GnCallback *selection_callback;
    Boolean call_callbacks;
};

class GnSingleSelection : public GnSelection {
    META_DEF_1(GnSingleSelection,GnSelection);
  public:
    GnSingleSelection(Boolean p_allow_no_selection = True,
		      Boolean p_call_immediately   = False);
  public:    
    // Return the selected item
    GnPresentable *SelectedItem();
    Boolean SelectionAvailable()
	{ return current != 0; }
    
    // Return the previously selected item
    GnPresentable *History();
    Boolean HistoryAvailable()
	{ return history != 0; }
    
    Boolean AllowNoSelection()
	{ return allow_no_selection; }
  protected:
    virtual void SelectItem(GnSelectionItem *, Boolean, Boolean);

    void change_value(void *client_data, caddr_t call_data);
    virtual void selection_changed();
    virtual void Removed(GnSelectionItem *p_item);

    virtual void CheckSelectedItems();
  protected:
    GnSelectionItem *history;
    GnSelectionItem *current;
    Boolean          allow_no_selection;
    Boolean          call_immediately;
};

class GnMultipleSelection : public GnSelection {
    META_DEF_1(GnMultipleSelection,GnSelection);
  public:
    GnMultipleSelection();
};

#endif










