// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnView.h   Version 1.36
//   LastSCCS    : 2/10/92  14:39:59
//   LastEdit    : "Mon Jan 20 14:27:07 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

#ifndef Gina_View_H
#define Gina_View_H

class GnDocument;
class GnCallback;
class GnScrolledWindow;

enum BUTTON_CODE { SELECT, EXTEND, MENU, TOGGLE, INSERT, OTHER };
/* = condensed form of...
     event->button = Button1        1
     event->button = Button2        2
     event->button = Button3        3
     event->button = Button4        4
     event->button = Button5        5
   and
     event->state = ShiftMask     1<<0      1
     event->state = LockMask      1<<1      2
     event->state = ControlMask   1<<2      4
     event->state = Mod1Mask      1<<3      8
     event->state = Mod2Mask      1<<4     16
     event->state = Mod3Mask      1<<5     32
     event->state = Mod4Mask      1<<6     64
     event->state = Mod5Mask      1<<7    128
     event->state = Button1Mask   1<<8    256
     event->state = Button2Mask   1<<9    512
     event->state = Button3Mask   1<<10  1024
     event->state = Button4Mask   1<<11  2048
     event->state = Button5Mask   1<<12  4096
*/

class GnView : public GnDrawingArea {
    META_DEF_1(GnView,GnDrawingArea);
    friend class GnViewObject;
  public:
    GnView( GnDocument *doc );
  public:
    GnDocument* get_document ()
	{ return document; }
    Boolean DrawingIsPossible();
    Window get_x_window()
	{ return x_window; }
    GC get_gcontext ()
	{ return gcontext; }
    GC get_feedback_gc ()
	{ return feedback_gc; };
    GnViewObjectList *get_view_objects()
	{ return(&view_objects); };
    Boolean overlapping_objects()
	{ return Overlapping_objects; }
    GnWidget *scroller() { return Parent(); }

    void PropagateButtonPress(Boolean  propagate)
	{ propagate_button_press = propagate; };
    Boolean PropagateButtonPress()
	{ return(propagate_button_press); };
    
    void receive_resize_callback(caddr_t call_data);
    void execute_expose_callback(void *client_data, caddr_t call_data);
    void execute_input_callback(void *client_data, caddr_t call_data);
    void handle_expose(caddr_t call_data);
    void handle_button_press(caddr_t call_data);
    void handle_key_press(caddr_t call_data);
    void handle_key_release(caddr_t call_data);

    /* Methods: */
    void define_cursor ();
    virtual void determine_window_attributes();

    void button_press_before( int code, int repetition, int x, int y );
    virtual void button_press( int code, int repetition, int x, int y );
    virtual void key_press( char key, int code, unsigned int modifiers,
			    int x, int y );
    virtual void key_release( char key, int code, unsigned int modifiers,
			     int x, int y );
    virtual void insert_selection( int x, int y );
    Boolean unselect_all_objects(); // Result True, if state changes
    void register_selection( GnViewObject *, Boolean register_obj =True );
    void clear_area( int x, int y, Dimension width, Dimension height );
    void force_redraw ();
    void invalidate_rectangle ( int, int, Dimension, Dimension );
    void execute_draw ( int count, int x, int y,
		        Dimension width, Dimension height );
    virtual void draw ( int count, int x, int y,
		        Dimension width, Dimension height );
    virtual void after_draw ( int count, int x, int y,
			      Dimension width, Dimension height );
    virtual void resized ();
    void print_gc ();
    void change_gc( XGCValues *val, unsigned long value_mask );
    void draw_into_pixmap ();
    void draw_point ( int x, int y );
    void draw_points ( XPoint *points, int npoints, Boolean relative =FALSE );
    void draw_line ( int x1, int y1, int x2, int y2, Boolean relative =FALSE );
    void draw_lines ( XPoint *points, int npoints, Boolean relative =FALSE,
		      Boolean fill =FALSE, int shape =Complex );
    void draw_segments ( XSegment *segments, int nsegments );
    void draw_rectangle ( int x, int y, Dimension width, Dimension height,
			  Boolean fill =FALSE );
    void draw_rectangles ( XRectangle *rects, int nrects, Boolean fill =FALSE );
    void draw_arc ( int x, int y, Dimension width, Dimension height,
		    int abgle1, int angle2, Boolean fill =FALSE );
    void draw_arcs ( XArc *arcs, int narcs, Boolean fill =FALSE );
    void draw_glyph ( int x, int y, char c );
    void draw_glyphs ( int x, int y, char* str );

  protected:  // regions
    static Boolean IntersectRectangles(int x1, int y1,
				       Dimension w1, Dimension h1,
				       int x2, int y2,
				       Dimension w2, Dimension h2);
    static Boolean PointInRectangle(int x1, int y1,
				    int x2, int y2,
				    Dimension w2, Dimension h2);
    void setGC_line_width( int width );
  protected:  // glyphs & strings
    void DrawGlyphs(GC gc, int x, int y, char* str, unsigned int length);
    void DrawGlyphs(int x, int y, char* str, unsigned int length)
	{ DrawGlyphs(default_text_gc, x, y, str, length); }

    void DrawStringAligned(GC gc, char *string, unsigned int length,
			   int x, int y, Dimension width, Dimension height,
			   unsigned char alignment = XmALIGNMENT_CENTER,
			   Dimension offset = 0);
    void DrawStringAligned(char *string, unsigned int length, int x, int y,
			   Dimension width, Dimension height,
			   unsigned char alignment = XmALIGNMENT_CENTER,
			   Dimension offset = 0)
	{ DrawStringAligned(default_text_gc, string, length,
			    x, y, width, height,
			    alignment, offset); }
  protected:

    virtual void before_create(GnWidget *);
    virtual void after_create();
    virtual Widget create_widget(GnWidget *, char *);
  protected:
    GnDocument *document;
    Window x_window;
    Boolean drawing_is_possible;
    Boolean backing_store;
    int repeated_click;
    Time last_click_time;
    Time double_click_intervall;
    Boolean printing;
  protected: // Fonts
    char *font_name;
    Font font;
    XmFontList fontlist;
    XFontStruct *font_info;
  protected: // graphics contexts
    GC gcontext;
    GC feedback_gc;
    GC default_text_gc;
    GC text_gc;
    Pixel foreground, background;
  protected: // view objects
    Boolean propagate_button_press;
    GnViewObjectList view_objects;
    Boolean Overlapping_objects;
    Boolean something_selected; // to be replaced by a list of selected objects
};

DefGenericListWithIterator(GnViewList,GnView,GnViewListIterator);

#endif Gina_View_H
