/* @(#)GnViewObject.h	1.20 1/23/92 */

#ifndef Gina_View_Object_H
#define Gina_View_Object_H

class GnView;


#define OBJ_NOTHING 0x0
#define OBJ_SELECTABLE 0x1
#define OBJ_MOVABLE 0x2
#define OBJ_RESIZABLE 0x4

class GnViewObject : public GnObject {
    META_DEF_1(GnViewObject,GnObject);
    friend ostream & operator<< ( ostream &, GnViewObject * );
    friend istream & operator>> ( istream &, GnViewObject * );
  protected:
    GnViewObject( XGCValues& cg_val, Dimension wid, Dimension hei,
		 Boolean mouse_sens = TRUE );
  public:

    /* Accessors: */
    int get_x_pos() { return x_pos; }
    int get_y_pos() { return y_pos; }
    GnView * get_parent_view() { return parent_view; }
    Dimension get_width() { return width; }
    Dimension get_height() { return height; }
    int facilities() { return(Facilities); }
    
    void select(Boolean redraw = True);
    void unselect(Boolean redraw = True);
    void toggle_selected(Boolean redraw = True);
    is_selected() { return(selected); };

    /* graphics context attributes: */
    void setGC_function( int );
    int getGC_function();
    void setGC_plane_mask( unsigned long );
    void setGC_foreground( unsigned long );
    void setGC_background( unsigned long );
    void setGC_line_width( int );
    int getGC_line_width();
    void setGC_line_style( int );
    void setGC_cap_style( int );
    void setGC_join_style( int );
    void setGC_fill_style( int );
    void setGC_fill_rule( int );
    void setGC_arc_mode( int );
    void setGC_tile( Pixmap );
    void setGC_stipple( Pixmap );
    void setGC_ts_x_origin( int );
    void setGC_ts_y_origin( int );
    void setGC_font( Font );
    void setGC_subwindow_mode( int );
    void setGC_graphics_exposures( Bool );
    void setGC_clip_x_origin( int );
    void setGC_clip_y_origin( int );
    void setGC_clip_mask( Pixmap );
    void setGC_dash_offset( int );
    void setGC_dashes( char );

    /* Methods: */
    Boolean is_mouse_sensitive() { return mouse_sensitive; }
    void install( GnView *, int x, int y, Boolean redraw =TRUE );
    void install( GnView *, Boolean redraw =TRUE );
    void deinstall ( Boolean redraw =TRUE );

    // Why is this a function of class GnViewObject ???
    inline Boolean point_inside_rectangle ( int x, int y,
					    int rect_x, int rect_y,
					    Dimension rect_width,
					    Dimension rect_height)
	{ return (x >= rect_x) && (y >= rect_y) &&
	         (x <= (rect_x + rect_width)) &&
		 (y <= (rect_y + rect_height)); }
    Boolean point_inside ( int x, int  y );
    virtual Boolean inside_rectangle ( int rect_x, int rect_y,
			       Dimension rect_width, Dimension rect_height );
    void move( int new_x, int new_y, Boolean redraw = TRUE );
    void resize( Dimension new_width, Dimension new_height,
		 Boolean redraw =TRUE );
    void invalidate_rectangle( int x, int y, Dimension width,
			       Dimension height );
    void force_redraw ();
    void execute_draw ( int count, int x, int y,
		        Dimension width, Dimension height );
    virtual void draw ( int count, int x, int y,
		        Dimension width, Dimension height );
    virtual void after_draw ( int count, int x, int y,
			      Dimension width, Dimension height );
    virtual void button_press ( int code, int repetition, int x, int y );
    virtual void became_selected();
    virtual void became_unselected();
    virtual void key_press ( int code, int x, int y, int root_x, int root_y );

    void draw_point ( int x, int y );
    void draw_points ( XPoint *points, int npoints, Boolean relative =FALSE );
    void draw_line ( int x1, int y1, int x2, int y2, Boolean relative =FALSE );
    void draw_lines ( XPoint *points, int npoints, Boolean relative =FALSE,
		      Boolean fill =FALSE, int shape =Complex );
    void draw_segments ( XSegment *segments, int nsegments );
    void draw_rectangle ( int x, int y, Dimension width, Dimension height,
			  Boolean fill =FALSE );
    void draw_rectangles ( XRectangle *rectangle, int nrectangles,
			   Boolean fill =FALSE );
    void draw_arc ( int x, int y, Dimension width, Dimension height,
		    int angle1, int angle2, Boolean fill =FALSE );
    void draw_arcs ( XArc *arcs, int narcs, Boolean fill =FALSE );
    void draw_glyph ( int x, int y, char c );
    void draw_glyphs ( int x, int y, char *string );
    
    virtual void write_to_stream( ostream &, Boolean FromSubclass = False);
    virtual void read_from_stream( istream &, Boolean FromSubclass = False);
    virtual int IO_type() { return(0); };
    
  protected:
    Dimension width;
    Dimension height;
    Boolean mouse_sensitive;
    int x_pos;
    int y_pos;
    GnView *parent_view;

    Boolean selected;
    int Facilities;

    XGCValues values;
    unsigned long value_mask;
};

DefGenericListWithIterator(GnViewObjectList,GnViewObject,GnViewObjectIterator);

#endif Gina_View_Object_H
