// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnWidget.h   Version 1.38
//   LastSCCS    : 2/10/92  14:49:30
//   LastEdit    : "Fri Feb  7 13:04:42 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven


#ifndef Gina_Widget_H
#define Gina_Widget_H

// Define some additional stuff for debugging widgets
#ifdef GINA_DEBUG
  inline void Gina_Debug_Proc_NL(Widget p)
    { cerr << "(Widget)" << (void*)p << endl; }
  inline void Gina_Debug_Proc(Widget p) { cerr << "(Widget)" << (void*)p; }
#endif GINA_DEBUG

class GnShell;
class GnToolkit;
class GnWidgetTable;
class GnGraphView;
class GnFormDialog;

typedef struct _GargList GargList, *GargListPtr;

class GnWidget : public GnObject {
    META_DEF_1(GnWidget,GnObject);
    friend class GnApplicationShell;
  public:
    // Constructors
    GnWidget();
    virtual ~GnWidget();

    // Preferred managed state at creation time
    virtual Boolean preferred_managed()
	{ return True; }

    // Creators
    GnWidget &create(GnWidget &parent, char * name)
	{ return *(create(&parent, name, preferred_managed())); }

    GnWidget *create(GnWidget *parent, char * name)
	{ return create(parent, name, preferred_managed()); }

    GnWidget &create(GnWidget &parent, char * name, Boolean managed)
	{ return *(create(&parent, name, managed)); }

    GnWidget *create(GnWidget *parent, char * name, Boolean managed)
	{ return _create(parent, name, managed); }

    // Virtual function to encapsulate an existing Xt Widget object
    virtual void encapsulate_self(Widget existing_widget);

    // Flush argument vectors of widget (and subwidgets in derived classes)
    virtual void SetValues();

    // Test if Motif widget has already been created
    Boolean IsMotifWidgetCreated()
	{ return widget_id != 0; }

    // Set/Get arguments by name
    void set_arg(char *, XtArgVal);
    void set_xmstring_arg(char *, char *);
    void set_font_arg(char *, char * const);
    void set_pixmap_arg(char *, char * const);
    void set_pixmap_arg(char *, char *, unsigned int, unsigned int);
    void get_arg(char *, XtArgVal );

    void realize();
    void manage();
    void unmanage();
    void Realize();
    void Manage();
    void Unmanage();
    Boolean IsManaged();
    void destroy();
    void move(Position new_x, Position new_y);
    void resize(Dimension new_width, Dimension new_height);

    void add_event_handler(GnCallback *, EventMask, Boolean = False);
    void remove_event_handler(GnCallback *, Boolean = True);

    // Methods
    void beep( int beeps = 1, int percent = 50 );

    // Accessors
    char *get_name()
	{ return XtName(widget_id); };
    Display *get_display()
	{ return XtDisplay(widget_id); };

    static GnToolkit *TheToolkit()
	{ return toolkit; };
    static GnWidgetTable *WidgetTable();
    
    void add_to_tab_group();
    void remove_from_tab_group();

    CALLBACK_DEF(destroyCallback);

    DEF_SET_GET_I(accelerators, XtAccelerators);
    DEF_SET_GET_I(ancestorSensitive, Boolean);
    DEF_SET_GET_I(background, Pixel);
    DEF_PIXMAP(backgroundPixmap);
    DEF_SET_GET_I(borderColor, Pixel);
    DEF_PIXMAP(borderPixmap);
    DEF_SET_GET_I(borderWidth, Dimension);
    DEF_SET_GET_I(colormap, Colormap);
    DEF_SET_GET_I(depth, int);
    DEF_SET(height, Dimension)
	{ get_group_widget()->set_arg(XmNheight, (XtArgVal)value); }
    DEF_GET(height, Dimension)
	{ Dimension p; get_group_widget()->get_arg(XmNheight, (XtArgVal)&p);
	  return p; }
    DEF_SET_GET_I(initialResourcesPersistent, Boolean);
    DEF_SET_GET_I(mappedWhenManaged, Boolean);
    DEF_SET_GET_I(screen, Screen *);
    DEF_SET_GET_I(sensitive, Boolean);
    DEF_SET_GET_I(translations, XtTranslations);
    DEF_SET(width, Dimension)
	{ get_group_widget()->set_arg(XmNwidth, (XtArgVal)value); }
    DEF_GET(width, Dimension)
	{ Dimension p; get_group_widget()->get_arg(XmNwidth, (XtArgVal)&p);
	  return p; }
    DEF_SET(x, Position)
	{ get_group_widget()->set_arg(XmNx, (XtArgVal)value); }
    DEF_GET(x, Position)
	{ Position p; get_group_widget()->get_arg(XmNx, (XtArgVal)&p);
	  return p; }
    DEF_SET(y, Position)
	{ get_group_widget()->set_arg(XmNy, (XtArgVal)value); }
    DEF_GET(y, Position)
	{ Position p; get_group_widget()->get_arg(XmNy, (XtArgVal)&p);
	  return p; }
    
    // XmPanedWindow Constraints

    DEF_SET(allowResize, Boolean)
	{ get_group_widget()->set_arg(XmNallowResize, (XtArgVal)value); }
    DEF_SET(maximum, int)
	{ get_group_widget()->set_arg(XmNmaximum, (XtArgVal)value); }

    DEF_SET(minimum, int)
	{ get_group_widget()->set_arg(XmNminimum, (XtArgVal)value); }
    DEF_SET(skipAdjust, Boolean)
	{ get_group_widget()->set_arg(XmNskipAdjust, (XtArgVal)value); }

    // XmForm Constraints: 

    DEF_SET(bottomAttachment, unsigned char)
	{ get_group_widget()->set_arg(XmNbottomAttachment, (XtArgVal)value); }
    DEF_SET(bottomOffset, int)
	{ get_group_widget()->set_arg(XmNbottomOffset, (XtArgVal)value); }
    DEF_SET(bottomPosition, int)
	{ get_group_widget()->set_arg(XmNbottomPosition, (XtArgVal)value); }
    DEF_SET(bottomWidget, GnWidget *)
	{ get_group_widget()->set_arg(XmNbottomWidget,
				      (XtArgVal)value->get_group_widget_id());};
    DEF_SET(bottomWidget, GnWidget &)
	{ get_group_widget()->set_arg(XmNbottomWidget,
				      (XtArgVal)value.get_group_widget_id()); };
    DEF_SET(leftAttachment, unsigned char)
	{ get_group_widget()->set_arg(XmNleftAttachment, (XtArgVal)value); }
    DEF_SET(leftOffset, int)
	{ get_group_widget()->set_arg(XmNleftOffset, (XtArgVal)value); }
    DEF_SET(leftPosition, int)
	{ get_group_widget()->set_arg(XmNleftPosition, (XtArgVal)value); }
    DEF_SET(leftWidget, GnWidget *)
	{ get_group_widget()->set_arg(XmNleftWidget,
				      (XtArgVal)value->get_group_widget_id());};
    DEF_SET(leftWidget, GnWidget &)
	{ get_group_widget()->set_arg(XmNleftWidget,
				      (XtArgVal)value.get_group_widget_id()); };
    DEF_SET(resizable, Boolean)
	{ get_group_widget()->set_arg(XmNresizable, (XtArgVal)value); }
    DEF_SET(rightAttachment, unsigned char)
	{ get_group_widget()->set_arg(XmNrightAttachment, (XtArgVal)value); }
    DEF_SET(rightOffset, int)
	{ get_group_widget()->set_arg(XmNrightOffset, (XtArgVal)value); }
    DEF_SET(rightPosition, int)
	{ get_group_widget()->set_arg(XmNrightPosition, (XtArgVal)value); }
    DEF_SET(rightWidget, GnWidget *)
 	{ get_group_widget()->set_arg(XmNrightWidget,
				      (XtArgVal)value->get_group_widget_id());};
    DEF_SET(rightWidget, GnWidget &)
 	{ get_group_widget()->set_arg(XmNrightWidget,
				      (XtArgVal)value.get_group_widget_id()); };
    DEF_SET(topAttachment, unsigned char)
	{ get_group_widget()->set_arg(XmNtopAttachment, (XtArgVal)value); }
    DEF_SET(topOffset, int)
	{ get_group_widget()->set_arg(XmNtopOffset, (XtArgVal)value); }
    DEF_SET(topPosition, int)
	{ get_group_widget()->set_arg(XmNtopPosition, (XtArgVal)value); }
    DEF_SET(topWidget, GnWidget *)
 	{ get_group_widget()->set_arg(XmNtopWidget,
				      (XtArgVal)value->get_group_widget_id());};
    DEF_SET(topWidget, GnWidget &)
 	{ get_group_widget()->set_arg(XmNtopWidget,
				      (XtArgVal)value.get_group_widget_id()); };

    // These functions should be protected and exportetd only to class GnWidget

    virtual Widget get_widget_id ();
    virtual GnWidget *get_group_widget ();
    virtual GnWidget *get_widget_for_sons ();

    Widget get_group_widget_id();
    Widget get_widget_id_for_sons();
    
    GnWidget *Parent();
    GnShell  *enclosing_shell();

    void print_widget();
    void dump_tree(int indent = 0);
    
    // virtual functions to support auto-scrolling
    virtual Boolean show_point( int, int ) {return(True);};
    virtual Boolean point_visible ( int, int ) {return(True);};
    virtual void scroll_to ( int, int ) {};
    virtual void register_scrolled_child( GnWidget *child );

  protected:
    // Virtual function to create the Motif widget
    virtual Widget create_widget(GnWidget *, char *);

    virtual GnWidget *_create(GnWidget *parent, char * name, Boolean managed);

    // Virtual function to check for a valid parent
    virtual Boolean check_parent(GnWidget *par)
	{ return( par != 0 ); };

    // This function is called before the widget is created
    virtual void before_create(GnWidget *)
	{};

    // This function is called after creating the widget but before managing
    virtual void after_create()
	{};

    // This function is called after the child has been managed
    virtual void after_manage()
	{};

    // Notify parent that a child has been created
    virtual void son_created(GnWidget *);
    
    void add_callback(GnCallback *callback, char *name);
    void remove_callback(GnCallback *callback, Boolean free_it);

    void reset_args();
    void adjust_args();

    Widget   widget_id;
    
    int args_n;		// Current number of arguments
    int args_max;       // Current maximum size of Arg vector
    Arg *args;	        // Argument vector

  private:
    static GnToolkit    *toolkit;
    static GnGraphView  *widget_tree_view;
    static GnFormDialog *widget_tree_dialog;
};

#endif Gina_Widget_H
