// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnCallback.C   Version 1.19
//   LastSCCS    : 1/23/92  12:33:16
//   LastEdit    : "Mon Jan 20 12:57:39 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

extern "C" {
    extern void XmAddProtocolCallback(Widget, Atom, Atom, XtCallbackProc,
				      XtPointer);
    extern void XmRemoveProtocolCallback(Widget, Atom, Atom, XtCallbackProc,
					 XtPointer);
}

META_IMPL_1(GnCallback,GnObject,);
ImplGenericListWithIterator(GnCallbackList,GnCallbackListIterator);

GnCallback::
GnCallback(GinaSimpleCallbackProc proc)
{
    function_name = proc;
    cb_type = CB_NOT_ATTACHED;
}

GnCallback::
~GnCallback()
{
    if( cb_type != CB_NOT_ATTACHED )
	remove();
}

void GnCallback::
evaluate(caddr_t call_data)
{
    (function_name)(call_data);
}

void GnCallback::
attach_to_widget(Widget widget, char *name)
{
    Require( cb_type == CB_NOT_ATTACHED );

    cb_type = CB_WIDGET;
    cb_info.callback.widget = widget;
    cb_info.callback.name = name;
    XtAddCallback(widget, name,
		  (XtCallbackProc)GnCallback::widget_handler,
		  (XtPointer)this);
}

void GnCallback::
attach_to_widget(Widget widget, EventMask mask, Boolean non_maskable)
{
    Require( cb_type == CB_NOT_ATTACHED );

    cb_type = CB_EVENT;
    cb_info.atom.widget = widget;
    cb_info.event.mask = mask;
    cb_info.event.non_maskable = non_maskable;
    XtAddEventHandler(widget, mask, non_maskable,
		      (XtEventHandler)GnCallback::widget_handler,
		      (XtPointer)this);
}

void GnCallback::
attach_to_widget(Widget widget, Atom property, Atom protocol)
{
    Require( cb_type == CB_NOT_ATTACHED );

    cb_type = CB_PROTOCOL;
    cb_info.atom.widget = widget;
    cb_info.atom.property = property;
    cb_info.atom.protocol = protocol;
    XmAddProtocolCallback(widget, property, protocol,
			  (XtCallbackProc)GnCallback::widget_handler,
			  (XtPointer)this);
}

void GnCallback::
attach_to_timer(GnTimer *timer)
{
    Require( cb_type == CB_NOT_ATTACHED );

    cb_type = CB_TIMER;
    cb_info.timer.timer = timer;
}

void GnCallback::
remove()
{
    Require( cb_type != CB_NOT_ATTACHED );

    switch(cb_type) {
      case CB_WIDGET:
	XtRemoveCallback(cb_info.callback.widget, cb_info.callback.name,
			 (XtCallbackProc)function_name, (XtPointer)this);
	break;
      case CB_EVENT:
	XtRemoveEventHandler(cb_info.event.widget, cb_info.event.mask,
			     cb_info.event.non_maskable,
			     (XtEventHandler)function_name, (XtPointer)this);
	break;
      case CB_PROTOCOL:
	XmRemoveProtocolCallback(cb_info.atom.widget,
				 cb_info.atom.property,
				 cb_info.atom.protocol,
				 (XtCallbackProc)function_name,
				 (XtPointer)this);
	break;
      case CB_TIMER:
	cb_info.timer.timer->remove_callback(this);
	break;
      case CB_NOT_ATTACHED:
	break;
    }
    cb_type = CB_NOT_ATTACHED;
}

void GnCallback::
widget_handler(Widget, GnCallback *cb, caddr_t call_data)
{
    cb->evaluate(call_data);
}

void GnCallback::
timer_handler(GnTimer *timer, XtIntervalId *)
{
    REQUIRE( ! timer->destroyed, "Timer is not destroyed" );
    REQUIRE(timer->enabled, "Timer is enabled" );
    REQUIRE(timer->running, "Timer is running" );

    //Gina_Debug_NL ("TimerDispatcher");

    timer->enabled = False; // TimeOut is removed automatically by Xt
    
    if( timer->active_in_handler )
	timer->install_timeout();
    
    timer->busy = True;
    for( GnCallbackListIterator it(timer->callbacks); ! it.Off(); it.Forth() )
	it.Item()->evaluate((caddr_t)(void *)timer);
    timer->busy = False;
    
    if(! timer->active_in_handler && timer->running &&
       ! timer->enabled && ! timer->destroyed )
	timer->install_timeout();
    
    if( timer->running  && timer->destroyed )
	timer->deinstall_timeout();
    
    if( timer->destroyed )
	delete timer;
}

META_IMPL_1(GnClassCallback,GnCallback,);

GnClassCallback :: GnClassCallback (GinaClassCallbackProc proc, void* obj)
: GnCallback((GinaSimpleCallbackProc)proc)
{
    object = obj;
}

void GnClassCallback::evaluate(caddr_t call_data)
{
    ((GinaClassCallbackProc)function_name)(object, call_data);
}

META_IMPL_1(GnCallbackOne,GnClassCallback,);

GnCallbackOne::
GnCallbackOne(GinaClassCallbackProcOne proc, void* obj, void *cd_one)
: GnClassCallback((GinaClassCallbackProc)proc, obj)
{
    client_data_one = cd_one;
}

void GnCallbackOne::evaluate(caddr_t call_data)
{
    ((GinaClassCallbackProcOne)function_name)(object, client_data_one,
					      call_data);
}

META_IMPL_1(GnCallbackTwo,GnCallbackOne,);

GnCallbackTwo::
GnCallbackTwo(GinaClassCallbackProcTwo proc, void* obj, void *cd1, void *cd2)
: GnCallbackOne((GinaClassCallbackProcOne)proc, obj, cd1)
{
    client_data_two = cd2;
}

void GnCallbackTwo::evaluate(caddr_t call_data)
{
    ((GinaClassCallbackProcTwo)function_name)(object,
					      client_data_one,
					      client_data_two,
					      call_data);
}
