// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnSelection.C   Version 1.10
//   LastSCCS    : 2/10/92  14:49:56
//   LastEdit    : "Wed Jan 22 13:08:41 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnSelection,GnObject,);

GnSelection::
GnSelection()
{
    selection_callback = 0;
    call_callbacks = True;
}    

GnSelectionItem *GnSelection::
FindSelectionItem(GnPresentable *p_item)
{
    for( GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth() )
	if( it.Item()->Value() == p_item )
	    return it.Item();
    return 0;
}

// ADDING OBJECTS

GnPresentable *GnSelection::
add(GnPresentable *obj, Boolean sel, Boolean sens)
{
    REQUIRE(obj != 0, "Non-zero object pointer");

    GnSelectionItem *item = create_item(obj, sel, sens);
    entries.Append(item);
    VisualizeItem(item);
    return obj;
}

GnPresentable *GnSelection::
add(GnPresentable &obj, Boolean sel, Boolean sens)
{
    return add(&obj, sel, sens);
}

GnPresentable *GnSelection::
add(char *str, Boolean sel, Boolean sens)
{
    REQUIRE( str != 0, "Not a NULL-pointer" );

    return add( new GnStringItem(str), sel, sens);
}

GnPresentable *GnSelection::
add(char *str, int i, Boolean sel, Boolean sens)
{
    REQUIRE( str != 0, "Not a NULL-pointer" );

    return add( new GnIntegerItem(str, i), sel, sens);
}

GnPresentable *GnSelection::
add(char *str, float f, Boolean sel, Boolean sens)
{
    REQUIRE( str != 0, "Not a NULL-pointer" );

    return add( new GnFloatItem(str, f), sel, sens);
}

GnPresentable *GnSelection::
add(int i, Boolean sel, Boolean sens)
{
    strstream str;
    str << i << '\0';
    CHECK( str.good(), "Stream is good" );

    return add( new GnIntegerItem(str.str(), i), sel, sens);
}

void GnSelection ::
add(GnPresentable *an_item_list[], int item_count, GnCallback *callback)
{
    REQUIRE(selection_callback == 0, "Callback is NULL" );

    for( int i = 0; i < item_count; i++ ) {
	GnSelectionItem *item = create_item(an_item_list[i], False, True);
	entries.Append(item);
	VisualizeItem(item);
    }
	
    selection_callback = callback;
}

void GnSelection ::
add(GnPresentableList *an_item_list, GnCallback *callback)
{
    Require(selection_callback == 0);

    GnPresentableListIterator iterator(an_item_list);

    for( iterator.Start(); ! iterator.Off(); iterator.Forth() ) {
	GnSelectionItem *item = create_item(iterator.Item(), False, True);
	entries.Append(item);
	VisualizeItem(item);
    }
    selection_callback = callback;
}

// INSERTING ITEMS

GnPresentable *GnSelection::
Insert(GnPresentable *p_item, GnPresentable *p_before,
       Boolean p_selected, Boolean p_sensitive)
{
    Require(False);
    return 0;
}

// REMOVING ITEMS

GnPresentable *GnSelection::
Remove(GnPresentable *p_item)
{
    for( GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth() ) {
	if( it.Item()->Value() == p_item ) {
	    GnSelectionItem *sel_item = it.Item();
	    UnvisualizeItem(sel_item);
	    it.Remove();
	    Removed(sel_item);
	    delete sel_item;
	    return p_item;
	}
    }
    return 0;
}

void GnSelection::
Removed(GnSelectionItem *p_item)
{
}

// CHANGING THE COMPLETE ITEM LIST

void GnSelection::
SetItemList(GnPresentable *new_item_list[], int new_item_count)
{
    UnvisualizeAllItems();
    for( GnSelectionItemListIterator sit(entries); ! sit.Off(); sit.Forth() ) {
	delete sit.Item();
    }
    entries.Clear();
    for( int i = 0; i < new_item_count; i++ )
	entries.Append(create_item(new_item_list[i], False, True));
    VisualizeAllItems();
    CheckSelectedItems();
}

void GnSelection::
SetItemList(GnPresentableList *new_item_list)
{
    UnvisualizeAllItems();
    for( GnSelectionItemListIterator sit(entries); ! sit.Off(); sit.Forth() ) {
	delete sit.Item();
    }
    entries.Clear();
    for( GnPresentableListIterator pit(new_item_list); ! pit.Off(); pit.Forth() )
	entries.Append(create_item(pit.Item(), False, True));
    VisualizeAllItems();
    CheckSelectedItems();
}

// ADD CALLBACKS

void GnSelection::
add_selectionCallback(GnCallback *callback)
{
    Require( callback != 0 );
    Require( selection_callback == 0 );

    selection_callback = callback;
}

// REMOVE CALLBACKS

void GnSelection::
remove_selectionCallback(GnCallback *callback, Boolean free_it)
{
    Require( callback != 0 );
    Require( selection_callback == callback );

    selection_callback = 0;
    if( free_it )
	delete callback;
}

// GET ITERATORS TO ITEMS

GnPresentableList *GnSelection::
AllItems()
{
    GnPresentableList *object_list = new GnPresentableList;
    for(GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth())
	object_list->Append(it.Item()->Value());
    return object_list;
}

GnPresentableList *GnSelection::
SelectedItems()
{
    GnPresentableList *object_list = new GnPresentableList;
    for(GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth())
	if(it.Item()->Selected())
	    object_list->Append(it.Item()->Value());
    return object_list;
}

GnPresentableList *GnSelection::
UnselectedItems()
{
    GnPresentableList *object_list = new GnPresentableList;
    for(GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth())
	if(! it.Item()->Selected())
	    object_list->Append(it.Item()->Value());
    return object_list;
}

GnPresentableList *GnSelection::
SensitiveItems()
{
    GnPresentableList *object_list = new GnPresentableList;
    for(GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth())
	if(it.Item()->Sensitive())
	    object_list->Append(it.Item()->Value());
    return object_list;
}

GnPresentableList *GnSelection::
InsensitiveItems()
{
    GnPresentableList *object_list = new GnPresentableList;
    for(GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth())
	if(! it.Item()->Sensitive())
	    object_list->Append(it.Item()->Value());
    return object_list;
}

// POSITION OF An ITEM

int GnSelection::
Position(GnPresentable *obj)
{
    GnSelectionItemListIterator iterator(entries);
    int pos = 0;
    
    for( iterator.Start(); ! iterator.Off(); iterator.Forth(), pos++ )
	if( iterator.Item()->Value() == obj )
	    return pos;
    return -1;
}

// QUERY AN ITEM

Boolean GnSelection::
IsSelected(GnPresentable *obj)
{
    GnSelectionItemListIterator iterator(entries);
    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    return(iterator.Item()->Selected());
    return False;
}

Boolean GnSelection::
IsSelected(GnPresentable &obj)
{
    return IsSelected(&obj);
}

Boolean GnSelection::
IsSensitive(GnPresentable *obj)
{
    GnSelectionItemListIterator iterator(entries);
    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    return(iterator.Item()->Sensitive());
    return False;
}

Boolean GnSelection::
IsSensitive(GnPresentable &obj)
{
    return IsSensitive(&obj);
}

// SELECT/UNSELECT ITEMS

void GnSelection::
Select(GnPresentable *obj, Boolean notify)
{
    GnSelectionItemListIterator iterator(entries);

    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    SelectItem(iterator.Item(), True, notify);
}

void GnSelection::
Unselect(GnPresentable *obj, Boolean notify)
{
    GnSelectionItemListIterator iterator(&entries);

    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    SelectItem(iterator.Item(), False, notify);
}

// CHANGE SENSITIVITY

void GnSelection::
Sensitive(GnPresentable *obj)
{
    GnSelectionItemListIterator iterator(&entries);

    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    SensitiveItem(iterator.Item(), True);
}

void GnSelection::
Insensitive(GnPresentable *obj)
{
    GnSelectionItemListIterator iterator(&entries);

    for(iterator.Start(); ! iterator.Off(); iterator.Forth() )
	if( iterator.Item()->Value() == obj )
	    SensitiveItem(iterator.Item(), False);
}

int GnSelection::
NumberOfSelectedItems(GnSelectionItem **p_first_selected_item)
{
    GnSelectionItemListIterator iterator(entries);
    int Result;

    if( p_first_selected_item )
	*p_first_selected_item = 0;

    for(Result = 0, iterator.Finish(); ! iterator.Off(); iterator.Back() )
	if( iterator.Item()->Selected() ) {
	    Result++;
	    if( p_first_selected_item )
		*p_first_selected_item = iterator.Item();
	}

    ENSURE_IF(p_first_selected_item && Result > 0, *p_first_selected_item != 0,
	      "Address of the first selected item is returned");
    ENSURE_IF(p_first_selected_item && Result == 0, *p_first_selected_item == 0,
	      "Address is NULL if no item is selected");

    return Result;
}

int GnSelection::
NumberOfSensitiveItems(GnSelectionItem **p_first_sensitive_item)
{
    GnSelectionItemListIterator iterator(entries);
    int n_sensitive;
    
    for(n_sensitive = 0, iterator.Finish(); ! iterator.Off(); iterator.Back())
	if( iterator.Item()->Sensitive() ) {
	    n_sensitive++;
	    if( p_first_sensitive_item )
		*p_first_sensitive_item = iterator.Item();
	}

    return n_sensitive;
}

// MAINTAINING ITEMS INTERNALLY

GnSelectionItem *GnSelection::
create_item(GnPresentable *obj, Boolean sel, Boolean sens)
{
    return new GnSelectionItem(obj, sel, sens);
}

void GnSelection::
VisualizeItem(GnSelectionItem *)
{
}

void GnSelection::
UnvisualizeItem(GnSelectionItem *)
{
}

void GnSelection::
SelectItem(GnSelectionItem *item, Boolean state, Boolean notify)
{
    Require(item->selectable());
    
    item->Select(state);
    call_callbacks = False;
    SelectItemWidget(item, state, notify);
    call_callbacks = True;

    if( notify )
	call_selection_callback();
}
    
void GnSelection::
SelectItemWidget(GnSelectionItem *, Boolean, Boolean)
{
}
    
void GnSelection::
SensitiveItem(GnSelectionItem *item, Boolean state)
{
    Require(item->sensitivity());
    
    item->Sensitive(state);
    SensitiveItemWidget(item, state);
}
    
void GnSelection::
SensitiveItemWidget(GnSelectionItem *, Boolean)
{
}
    
void GnSelection::
VisualizeAllItems()
{
    for( GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth() )
	VisualizeItem(it.Item());
}

void GnSelection::
UnvisualizeAllItems()
{
    for( GnSelectionItemListIterator it(entries); ! it.Off(); it.Forth() )
	UnvisualizeItem(it.Item());
}

void GnSelection::
CheckSelectedItems()
{
}

void GnSelection::
change_value(void *, caddr_t)
{
    if( call_callbacks )
	call_selection_callback();
}

void GnSelection::
call_selection_callback()
{
    if( selection_callback )
	selection_callback->evaluate((caddr_t)(GnSelection *)this);
    else
	selection_changed();
}

void GnSelection::
selection_changed()
{
    GnSelectionItemListIterator it(entries);
    char str[50];
    DumpOn(cout);
    cout << " with " << entries.Length() << " items" << endl;
    for(it.Start(); ! it.Off(); it.Forth() ) {
	cout << it.Item()->Value()->PrintOn(str, 50) << " item: ";
	it.Item()->DumpOn(cout);
	cout << " addr: " << (void *)it.Item()->Value();
	cout << " Sel: " << (int)it.Item()->Selected();
	cout << " Sens: " << (int)it.Item()->Sensitive() << endl;
    }
}


META_IMPL_1(GnSingleSelection,GnSelection,);

GnSingleSelection::
GnSingleSelection(Boolean p_allow_no_selection, Boolean p_call_immediately)
{
    history = current = 0;
    allow_no_selection = p_allow_no_selection;
    call_immediately = p_call_immediately;
}    

GnPresentable *GnSingleSelection::
SelectedItem()
{
    REQUIRE( SelectionAvailable(), "Selection is available" );
    
    return current->Value();
}
    
GnPresentable *GnSingleSelection::
History()
{
    REQUIRE( HistoryAvailable(), "History is available" );
    
    return history->Value();
}

void GnSingleSelection::
SelectItem(GnSelectionItem *item, Boolean state, Boolean notify)
{
    REQUIRE(item->selectable(), "Item is selectable");
    REQUIRE(state == True, "No deselection from outside !!" );

    // Update history
    history = current;

    // (De)select current item WITHOUT calling callbacks
    if( current != 0 && current != item) {
	call_callbacks = False;
	current->Select(False);
	SelectItemWidget(item, False, False);
	call_callbacks = True;
	current = item;
    }
    // Select the new item w/o notification
    GnSelection::SelectItem(item, state, call_immediately ? notify : False );
}
    
void GnSingleSelection::
Removed(GnSelectionItem *p_item)
{
    REQUIRE( p_item != 0, "Not NULL" );

    if( history == p_item )
	history = 0;
    if( current == p_item )
	current = 0;

    ENSURE( current != 0 || AllowNoSelection(), "");
}

void GnSingleSelection::
CheckSelectedItems()
{
    REQUIRE_IF( ! AllowNoSelection(),
	        ! entries.Empty(), "There is at least one item for selection");

    GnSelectionItem *selected_item = 0;
    int n_selected = NumberOfSelectedItems(&selected_item);

    REQUIRE( n_selected <= 1, "Not more than one element is selected");

    history = current = selected_item;

    if( n_selected == 0 && ! AllowNoSelection() ) {
	GnSelectionItem *sensitive_item;
	int n_sensitive = NumberOfSensitiveItems(&sensitive_item);

	REQUIRE( n_sensitive >= 1, "At least one element is sensitive");

	sensitive_item->Select(True);
	history = current = sensitive_item;
    }

    ENSURE_IF(AllowNoSelection(), NumberOfSelectedItems() <= 1,
	      "At mots one element is selected");
    ENSURE_IF(! AllowNoSelection(), NumberOfSelectedItems() == 1,
	      "Exactly one element is selected");
}

void GnSingleSelection::
selection_changed()
{
    GnSelection::selection_changed();
    cout << "current = " 
         << (void *)(current ? current->Value() : 0)
         << " history = "
         << (void *)(history ? history->Value() : 0)
         << endl;
}

void GnSingleSelection::
change_value(void *client_data, caddr_t)
{
    // We expect the newly selected item on the client data argument
    
    GnSelectionItem *new_selection = (GnSelectionItem *)client_data;

    if( current != new_selection ) {
	history = current;
	call_callbacks = False;
	if( current )
	    current->Select(False);
	current = new_selection;
	current->Select(True);
	call_callbacks = True;
	call_selection_callback();
    }
}


META_IMPL_1(GnMultipleSelection,GnSelection,);

GnMultipleSelection::
GnMultipleSelection()
{
}

