// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnWidget.C   Version 1.44
//   LastSCCS    : 2/10/92  15:12:59
//   LastEdit    : "Mon Feb 10 12:52:57 1992"
//   Description : 
//   Author      : Andreas Baecker, Robert Babatz
//   Copyright   : GMD Schloss Birlinghoven

GnToolkit    *GnWidget::toolkit = 0;
GnGraphView  *GnWidget::widget_tree_view = 0;
GnFormDialog *widget_tree_dialog = 0;

META_IMPL_1(GnWidget,GnObject,);

GnWidget ::
GnWidget ()
{
    widget_id = 0;
    args_n    = 0;
    args_max  = 0;
    args      = 0;
}

GnWidget ::
~GnWidget ()
{
    // Gina_Debug("Destructor::GnWidget::this => ");
    // Gina_Debug_NL((void *)this);
    // Gina_Debug("Destructor::GnWidget::this (with cast) => ");
    // Gina_Debug_NL((void *)(GnWidget *)this);
    // Gina_Debug("Destructor::GnWidget::widget_id => ");
    // Gina_Debug_NL((void *)widget_id);

    // The widget_id may be 0 if the widget object is deleted immediately
    // after it has been created.
    if( widget_id != 0 )
        toolkit->widget_table.remove(this, widget_id);
}

GnWidgetTable *GnWidget::
WidgetTable()
{
    return &(toolkit->widget_table);
}

void GnWidget::reset_args()
{
    args_n = args_max = 0;
    if( args ) {
	XtFree((char *)args);
	args = 0;
    }
}

void GnWidget::adjust_args()
{
    if(args) {
	args_max += 10;
	args = (Arg *)XtRealloc((char *)args, sizeof(Arg)*args_max);
    } else {
	args_max = 20;
	args = (Arg *)XtMalloc(sizeof(Arg)*args_max);
    }
}

Widget GnWidget ::
get_widget_id () 
{
    return(widget_id);
}

GnWidget *GnWidget ::
get_group_widget ()
{
    return(this);
}

GnWidget *GnWidget ::
get_widget_for_sons ()
{
    return(this);
}


Widget GnWidget ::
get_group_widget_id()
{
    return get_group_widget()->get_widget_id();
}

Widget GnWidget ::
get_widget_id_for_sons()
{
    return get_widget_for_sons()->get_widget_id();
}
    
Widget GnWidget ::
create_widget(GnWidget *, char *)
{
    return(0);
}

void GnWidget::
son_created(GnWidget *)
{
}

GnWidget *GnWidget ::
_create(register GnWidget *parent, char * widget_name, Boolean managed)
{
    REQUIRE( ! IsMotifWidgetCreated(), "Widget is not created" );
    REQUIRE(check_parent(parent), "Parent is legal" );
    REQUIRE_IF(parent != 0, parent->IsMotifWidgetCreated(),
	       "Parent widget is created");

    before_create(parent);
    widget_id = create_widget(parent, widget_name);
    toolkit->widget_table.insert(this, widget_id);
    reset_args();
    after_create();
    if( parent )
	parent->son_created((GnWidget *)this);
    if( managed )
	XtManageChild(widget_id);
    after_manage();
    //print_widget();
    
    Ensure( widget_id != NULL );
    Ensure( args_n == 0 );
    
    return this;
}

GnWidget *GnWidget::
Parent()
{
    return toolkit->widget_table.lookup(XtParent(get_group_widget_id()));
}

void GnWidget ::
register_scrolled_child( GnWidget *)
{
}

void GnWidget ::
encapsulate_self(Widget widget)
{
    widget_id = widget;
    toolkit->widget_table.insert(this, widget);
    //print_widget();
}

void GnWidget ::
beep( int beeps, int percent )
{
    for ( int i = 0; i < beeps; i++ )
	XBell( get_display(), percent );
}

void GnWidget::add_callback(GnCallback *callback, char *callback_name)
{
    callback->attach_to_widget(get_widget_id(), callback_name);
}

void GnWidget::remove_callback(GnCallback *callback, Boolean free_it)
{
    callback->remove();
    if(free_it)
	delete callback;
}

void GnWidget::add_event_handler(GnCallback *event, EventMask mask,
				Boolean non_maskable)
{
    event->attach_to_widget(get_widget_id(), mask, non_maskable);
}

void GnWidget::remove_event_handler(GnCallback *event, Boolean free_it)
{
    event->remove();
    if(free_it)
	delete event;
}

void GnWidget ::
destroy ()
{
    XtDestroyWidget(get_group_widget_id());
}

void GnWidget::
realize()
{
    Realize();
}

void GnWidget ::
manage ()
{
   Manage();
}

void GnWidget ::
unmanage ()
{
    Unmanage();
}

void GnWidget::
Realize()
{
    //Gina_Debug("XtRealizeWidget=");
    //Gina_Debug_NL((void *)get_group_widget_id());
    XtRealizeWidget(get_group_widget_id());
}

void GnWidget ::
Manage ()
{
    XtManageChild( get_group_widget_id() );
}

void GnWidget ::
Unmanage ()
{
    XtUnmanageChild( get_group_widget_id() );
}

Boolean GnWidget ::
IsManaged()
{
    return XtIsManaged(get_group_widget_id());
}

void GnWidget ::
move( Position new_x, Position new_y )
{
    get_group_widget()->setR_x(new_x);
    get_group_widget()->setR_y(new_y);
    if( get_group_widget()->IsMotifWidgetCreated() )
	get_group_widget()->SetValues();
}

void GnWidget ::
resize( Dimension new_width, Dimension new_height )
{
    get_group_widget()->setR_width(new_width);
    get_group_widget()->setR_height(new_height);
    if( get_group_widget()->IsMotifWidgetCreated() )
	get_group_widget()->SetValues();
}

void GnWidget ::
add_to_tab_group ()
{
}

void GnWidget ::
remove_from_tab_group ()
{
}

GnShell * GnWidget ::
enclosing_shell ()
{
    return Parent()->enclosing_shell();
}

void GnWidget ::
SetValues ()
{
    if( get_group_widget() != (GnWidget *)this )
	get_group_widget()->SetValues();
    if( args_n > 0 ) {
	XtSetValues(get_widget_id(), args, args_n);
	reset_args();
    }
}

void GnWidget ::
set_arg (char * str, XtArgVal val)
{
    if( args_n >= args_max )
	adjust_args();
    XtSetArg (args[args_n], str, val); args_n++;
}

void GnWidget::
set_xmstring_arg(char *resource_name, char *str)
{
    set_arg(resource_name, (XtArgVal)(GnXMSTRING_LTOR(str)));
}

void GnWidget::
set_font_arg(char *resource_name, char * const font_name)
{
    XFontStruct *font;
    XmStringCharSet cset = (XmStringCharSet)XmSTRING_DEFAULT_CHARSET;

    if( (font = XLoadQueryFont(toolkit->AppDisplay(), font_name) ) == NULL ) {
	cerr << "GINA Warning: Cant't load font " << font_name << endl;
	cerr.flush();
    }
    set_arg(resource_name, (XtArgVal)XmFontListCreate(font, cset));
}

extern "C" {
    extern Pixmap XmGetPixmap(Screen *, char *, Pixel, Pixel);
}
			   
void GnWidget::
set_pixmap_arg(char *resource_name, char * const image_name)
{
    Screen *screen = toolkit->AppShell().getR_screen();
    //XtScreen(toolkit->AppShell().get_widget_id());
    Pixmap pixmap = XmGetPixmap(screen, image_name,
				XBlackPixelOfScreen(screen),
				XWhitePixelOfScreen(screen));
    if( pixmap ==  XmUNSPECIFIED_PIXMAP ) {
	cerr << "GINA Warning: Can't open Pixmap " << image_name << endl;
	cerr.flush();
    }
    set_arg(resource_name, (XtArgVal)pixmap);
}
    
void GnWidget::
set_pixmap_arg(char *resource_name, char *image_bits,
	       unsigned int width, unsigned int height)
{
    Screen *screen = toolkit->AppShell().getR_screen();
    
    Pixmap pixmap = XCreatePixmapFromBitmapData(toolkit->AppDisplay(),
						DefaultRootWindow
						  (toolkit->AppDisplay()),
						image_bits, width, height,
						XBlackPixelOfScreen(screen),
						XWhitePixelOfScreen(screen),
						XDefaultDepthOfScreen(screen));
    set_arg(resource_name, (XtArgVal)pixmap);
}
    
void GnWidget::
get_arg(char * resource_name, XtArgVal val_ptr)
{
    Arg arg[1];

    XtSetArg(arg[0], resource_name, val_ptr);
    XtGetValues(get_widget_id(), arg, 1);
}

void GnWidget ::
print_widget ()
{
    Gina_Debug(widget_id); Gina_Debug("; ");
    Gina_Debug("parent: "); Gina_Debug_NL(XtParent(widget_id));
}

void GnWidget::
dump_tree(int indent)
{
    for(int i =0; i < indent; i++)
	cout << " ";
    cout << get_name() << " Class: " << Class()->Name() << endl;;

    Arg args[2];
    WidgetList wl;
    int num_ch;
    GnWidget *new_w, *true_w;

    if( XtIsComposite(get_widget_id()) ) {
	XtSetArg(args[0], XmNchildren, &wl);
	XtSetArg(args[1], XmNnumChildren, &num_ch);
	XtGetValues(get_widget_id(), args, 2);

	for( int j = 0; j < num_ch; j++ ) {
	    new_w = new GnWidget;
	    true_w = toolkit->widget_table.lookup(wl[j], new_w, True);
	    true_w->dump_tree(indent+2);
	}
    }
    cout.flush();
}
